/*
  Copyright Eric Reese, 2005
  This code is part of KLBDT and is distributed under the GNU GPL, version 2 or
  (at your option) any later version as published by the Free Software Foundation.
  This comes with no warranty whatsoever, without even the implied warranty of
  merchantability or fitness for a particular purpose.  It might not even work.
  See the GNU GPL for more details. http://www.gnu.org/licenses/gpl.txt

  http://kyzentun.tejat.net/contact.html lists ways of contacting me.
 */
#include "draw.hh"
#include "error.hh"
#include "line.hh"

using std::istream;
using std::ostream;
using std::endl;
using std::vector;
using std::pair;
using std::vector;

basic_line::basic_line()
  {}

basic_line::basic_line(vector<point> pixies)
  :pixels(pixies)
  {}

bool line::is_valid() const
  {
    return has_been_set;
  }

line::line()
  :invisible(false)
   , interpolated(false)
   , hilighted(false)
   , s_color(0)
   , e_color(0)
   , has_been_set(false)
  {}

line::line(const line& foo)
  :basic_line(foo.pixels)
   , invisible(foo.invisible)
   , interpolated(foo.interpolated)
   , hilighted(foo.hilighted)
   , s_color(foo.s_color)
   , e_color(foo.e_color)
   , has_been_set(true)
  {}

void line::draw() const
  {
    if(!is_valid())
    {
      throw error("Invalid line.");
    }
    if(!invisible)
    {
      Uint32 current_color= s_color;
      for(vector<point>::const_iterator n= pixels.begin(); n != pixels.end(); ++n)
      {
        if(interpolated)
        {
          current_color= interpolate(s_color, current_color, e_color, pixels.size());
        }
        draw_pixel(*n, current_color, hilighted);
      }
    }
  }

void line::draw_point(point p) const
  { /* FIXME?  Should it check for invisibility, validity, and interpolation? */
    /* Also, should it check to make sure the point is on the line? */
    draw_pixel(p, s_color, hilighted);
  }

void line::draw_ignore_special() const
  {
    if(!is_valid())
    {
      throw error("Invalid line.");
    }
    for(vector<point>::const_iterator n= pixels.begin(); n != pixels.end(); ++n)
    {
      draw_pixel(*n, s_color, false);
    }
  }

void line::draw_invisible() const
  { /* Um, why? */
    if(!is_valid())
    {
      throw error("Invalid line.");
    }
  }

void line::draw_interpolated() const
  {
    if(!is_valid())
    {
      throw error("Invalid line.");
    }
    if(!invisible)
    {
      Uint32 current_color= s_color;
      for(vector<point>::const_iterator n= pixels.begin(); n != pixels.end(); ++n)
      {
        current_color= interpolate(s_color, current_color, e_color, pixels.size());
        draw_pixel(*n, current_color, hilighted);
      }
    }
  }

void line::draw_hilighted() const
  {
    if(!is_valid())
    {
      throw error("Invalid line.");
    }
    if(!invisible)
    {
      Uint32 current_color= s_color;
      for(vector<point>::const_iterator n= pixels.begin(); n != pixels.end(); ++n)
      {
        if(interpolated)
        {
          current_color= interpolate(s_color, current_color, e_color, pixels.size());
        }
        draw_pixel(*n, current_color, true);
      }
    }
  }

void line::draw_as_color(Uint32 color) const
  {
    if(!is_valid())
    {
      throw error("Invalid line.");
    }
    for(vector<point>::const_iterator n= pixels.begin(); n != pixels.end(); ++n)
    {
      draw_pixel(*n, color, hilighted);
    }
  }

void line::draw_special(bool d_invis, bool d_interp, bool d_hilight, Uint32 d_s_c
  , Uint32 d_e_c) const
  {
    if(!is_valid())
    {
      throw error("Invalid line.");
    }
    if(!d_invis)
    {
      Uint32 current_color= d_s_c;
      for(vector<point>::const_iterator n= pixels.begin(); n != pixels.end(); ++n)
      {
        if(d_interp)
        {
          current_color= interpolate(d_s_c, current_color, d_e_c, pixels.size());
        }
        draw_pixel(*n, current_color, d_hilight);
      }
    }
  }

void line::draw_really_special(pair<bool, bool> d_invis, pair<bool, bool> d_interp
  , pair<bool, bool> d_hilight, pair<bool, Uint32> d_s_c
  , pair<bool, Uint32> d_e_c) const
  {
    if(!is_valid())
    {
      throw error("Invalid line.");
    }
    if(!d_invis.first || !d_invis.second)
    {
      Uint32 current_color= (d_s_c.first) ? (d_s_c.second) : (s_color);
      for(vector<point>::const_iterator n= pixels.begin(); n != pixels.end(); ++n)
      {
        if(d_interp.first && d_interp.second)
        {
          current_color= interpolate(d_s_c.first ? d_s_c.second : s_color
            , current_color, d_e_c.first ? d_e_c.second : e_color, pixels.size());
        }
        draw_pixel(*n, current_color, d_hilight.first ? d_hilight.second : hilighted);
      }
    }
  }

void line::set_pixels(point start, basic_line& line_template, bool invis, bool interp
  , bool hilight, Uint32 s_c, Uint32 e_c)
  {
    pixels.clear();
    point current_pixel= start;
    pixels.push_back(current_pixel);
    for(vector<point>::iterator n= line_template.pixels.begin(); n != line_template.pixels.end(); ++n)
    {
      current_pixel+= *n;
      pixels.push_back(current_pixel);
    }
    invisible= invis;
    interpolated= interp;
    hilighted= hilight;
    s_color= s_c;
    e_color= e_c;
    has_been_set= true;
  }

Uint32 line::shift_channel(bool s_or_e, unsigned int channel, int amount)
  {
    if(!is_valid())
    {
      throw error("Invalid line.");
    }
    Uint32* color_to_manipulate;
    if(s_or_e)
      color_to_manipulate= &s_color;
    else
      color_to_manipulate= &e_color;
    channel &= 3; /* 0b11 */ /* I'm fairly sure bitwise and is faster than modulus.
              Also, this supports the possibility that SDL stores colors as ARGB. */
    if(amount > 255)
    amount= 255;
    if(amount < -255)
    amount= -255;
    unsigned int temp= (*color_to_manipulate) >> (8 * channel);
    temp&= 255;
    temp+= amount;
    if(temp > 255)
    {
      if(amount < 0)
      {
        temp= 0;
      }
      else
      {
        temp= 255;
      }
    }
    (*color_to_manipulate)= ((*color_to_manipulate) & (~(255 << (channel * 8))))
      | ((temp) << (channel * 8));
    return *color_to_manipulate;
  }

bool operator==(line rhs, line lhs)
  {
    if(rhs.pixels.size() != lhs.pixels.size())
    {
      return false;
    }
    for(vector<point>::size_type n= 0; n < lhs.pixels.size(); ++n)
    {
      if(lhs.pixels[n] != rhs.pixels[n])
      {
        return false;
      }
    }
    return (lhs.s_color == rhs.s_color) && (lhs.e_color == rhs.e_color)
      && (lhs.invisible == rhs.invisible) && (lhs.interpolated == rhs.interpolated)
      && (lhs.hilighted == rhs.hilighted);
  }

ostream& line::spew_info(ostream& out_file)
  {
    if(out_file)
    {
      if(!is_valid())
      {
        throw error("Invalid line.");
      }
      out_file << invisible << ' ' << interpolated << ' ' << hilighted << ' '
        << s_color << ' ' << e_color << ' ';
      for(vector<point>::iterator n= pixels.begin(); n != pixels.end(); ++n)
      {
        out_file << n->x << ' ' << n->y << ' ';
      }
      out_file << endl;
    }
    return out_file;
  }

istream& line::read_info(istream& in_file)
  {
    if(in_file)
    {
      if(in_file >> invisible)
      {
        if(in_file >> interpolated)
        {
          if(in_file >> hilighted)
          {
            if(in_file >> s_color)
            {
              if(in_file >> e_color)
              {
                for(vector<point>::iterator n= pixels.begin(); n != pixels.end(); ++n)
                {
                  if(!(in_file >> n->x) || !(in_file >> n->y))
                  {
                    break;
                  }
                }
              }
            }
          }
        }
      }
      if(in_file)
      {
        has_been_set= true;
      }
    }
    return in_file;
  }

// istream& line::read_unformatted_info(istream& in_file)
//   {
//     if(in_file)
//     {
//
//     }
//     return in_file;
//   }
