/*
  Copyright Eric Reese, 2005
  This code is part of KLBDT and is distributed under the GNU GPL, version 2 or
  (at your option) any later version as published by the Free Software Foundation.
  This comes with no warranty whatsoever, without even the implied warranty of
  merchantability or fitness for a particular purpose.  It might not even work.
  See the GNU GPL for more details. http://www.gnu.org/licenses/gpl.txt

  http://kyzentun.tejat.net/contact.html lists ways of contacting me.
 */
#ifndef LINE_SET_HH
#define LINE_SET_HH

#include <deque>
#include <iostream>
#include <list>
#include <map>
#include <utility>

#include "SDL.h"

#include "line.hh"

struct line_set
{
  typedef std::deque<line*> line_container_t;

  line_set()
  {}

  ~line_set();
  
  line* access_back_line();
  line* access_front_line();
  line* access_by_index(int index);
  /* Positive index is from the front, negative index is from the back. */
  line* access_by_point(point pixel_of_line) throw(error);
  /* Access lines in various ways. */
  
  void push_back_line(line& foo);
  void push_front_line(line& foo);
  void push_before_index(int index, line& foo);
  void push_at_point(point pos, line& foo) throw(error);
  /* Add a line in various ways.
     Assumes that the line is already set up correctly. */
  
  line pop_back_line();
  line pop_front_line();
  line pop_by_index(int index);
  line pop_by_point(point pixel_of_line) throw(error);
  /* A map of points to pointers to lines allows this to be reasonably fast. */
  /* The potential for missing the line is the reason why zooming is provided. */
  /* Remove a line from the front, back, or anywhere else. */

  void add_set(line_set& set_to_add);

  void mung();
  line_container_t::size_type size();
  bool empty() const;

  std::ostream& save_set_to_file(std::ostream& out_file);
  std::istream& load_set_from_file(std::istream& in_file) throw(error);

//   std::ostream& save_set_unformatted(std::ostream& out_file);
//   std::istream& load_set_unformatted(std::istream& in_file);

  void draw_all() const;
  void draw_all_ignore_global_specials() const;
  void draw_all_ignore_special() const;
  void draw_all_invisible() const; /* Silly, isn't it? */
  void draw_all_interpolated() const;
  void draw_all_hilighted() const;
  void draw_all_as_color(Uint32 color) const;
  void draw_all_special(bool d_invis, bool d_interp, bool d_hilight, Uint32 d_s_c
    , Uint32 d_e_c) const;
  void draw_all_really_special(std::pair<bool, bool> d_invis
    , std::pair<bool, bool> d_interp, std::pair<bool, bool> d_hilight
    , std::pair<bool, Uint32> d_s_c, std::pair<bool, Uint32> d_e_c) const;

  template<class Point_iterator>
    void redraw_points(Point_iterator beg, Point_iterator end)
  {
    for(Point_iterator p= beg; p != end; ++p)
    {
      std::map<point, line*>::iterator tmp= lines_by_point.find(*p);
      if(tmp == lines_by_point.end())
      {
        continue;
      }
      tmp->second->draw_point(*p);
    }
  }
  
  bool all_invis;
  bool all_interp;
  bool all_hilight;
  Uint32 all_s_color, all_e_color;
  bool use_all_invis, use_all_interp, use_all_hilight, use_all_s_color
    , use_all_e_color, use_all_alls;
  /* use_all_* specifies whether to use the associated all_* when drawing lines.
     Using one of these overrides the corresponding members of the lines.
     If use_all_* is false, then the associated all_* is ignored.
     If use_all_alls is true, then all all_* members are used, overriding the
       associated use_all_* members.
     If use_all_alls is false, then the individual use_all_* members are checked. */
private:
  /* Here's the plan:  Store the actual lines in a deque for fast additions/removals,
     and so that we don't have to implement push_front() and pop_front() for vector.
     Each time a line is created, make an entry in the map for each point it contains
     with the line* pointing to the line.  This gives us faster access by point, and
     same speed access by index. */
  /* As a side note, the new point access method means that newer lines get the point
     in overlap cases.  The old system gave the point to the older line. */
  /* Problem:  Adding a line that overlaps an old line takes that point away from the
     old line.  When the new line is removed, that point is not given back.  Thus, it
     is possible for a point to be on a line, but unusable for accessing that line. */
  /* Solution:  Store the overlapped points in a map of points to vectors of pointers
     of lines.  When adding points, check for overlap, and save old point and line*
     pair if necessary.  When removing points, check to see if old points exist, and
     only delete the entry in the map if no old points exist. */
  template<class Point_iterator>
    void add_points_to_map_with_line(Point_iterator beg, Point_iterator end, line* ref_to)
  {
    for(Point_iterator p= beg; p != end; ++p)
    {
      std::map<point, line*>::iterator tmp= lines_by_point.find(*p);
      if(tmp == lines_by_point.end())
      {
        lines_by_point[*p]= ref_to;
        continue;
      }
      overlapped_points[*p].push_back(tmp->second);
      tmp->second= ref_to;
    }
  }
  template<class Point_iterator>
    void remove_points_from_map(Point_iterator beg, Point_iterator end)
  {
    for(Point_iterator p= beg; p != end; ++p)
    {
      std::map<point, line*>::iterator l= lines_by_point.find(*p);
      if(l != lines_by_point.end())
      {
        std::map<point, std::list<line*> >::iterator s= overlapped_points.find(*p);
        if(s == overlapped_points.end())
        {
          lines_by_point.erase(l);
          continue;
        }
        l->second= s->second.back();
        s->second.pop_back();
        if(s->second.empty())
        {
          overlapped_points.erase(s);
        }
      }
    }
  }
  std::map<point, line*> lines_by_point;
  line_container_t actual_lines;
  std::map<point, std::list<line*> > overlapped_points;
};

#endif
