/*
  Copyright Eric Reese, 2005
  This code is part of KLBDT and is distributed under the GNU GPL, version 2 or
  (at your option) any later version as published by the Free Software Foundation.
  This comes with no warranty whatsoever, without even the implied warranty of
  merchantability or fitness for a particular purpose.  It might not even work.
  See the GNU GPL for more details. http://www.gnu.org/licenses/gpl.txt

  http://kyzentun.tejat.net/contact.html lists ways of contacting me.
 */
/*  You're probably wondering why I wrote this program, when there are great image
      editors like The Gimp out there.
    In my drawings, I combine a few simple operations, each requiring great attention
      to detail.  With the Gimp, these operations turned out to be rather tedious.
    So I wrote this tool to perform those operations for me to save time on those parts.
    I've estimated that I can create a large drawing from scratch at least 8 times
      faster with this tool than with The Gimp.
    Yes, I know The Gimp has script-fu, but I'd still want to write a specialized UI,
      and I'd rather do that in a language I'm more familiar with.
*/
// #include <algorithm>
#include <cctype>
#include <fstream>
#include <iostream>
#include <list>
#include <map>
#include <sstream>
#include <string>
// #include <utility>
#include <vector>

#include "SDL.h"
#include "SDL_nfnt.h"

#include "draw.hh"
#include "environment.hh"
#include "error.hh"
#include "sdl_command_line.hh"

using std::ifstream;
using std::ofstream;
using std::istream;
using std::cout;
using std::cerr;
using std::endl;
using std::list;
using std::map;
using std::string;
using std::istringstream;
using std::ostringstream;
using std::vector;

void show_usage(string prog_name);
int init_everything(list<string>& template_files);

Uint32 starting_line_color= 0xffffff;
Uint32 current_line_color= starting_line_color;

int init_everything(list<string>& template_files)
  {
    try
    {
      init_drawing();
      current_env.init(template_files);
    }
    catch(error e)
    {
      cerr << "Error in initialization:  " << e.cause << endl;
      return -1;
    }
    cerr << "Success." << endl;
    return 0;
  }

void show_usage(string prog_name)
  {
    cout << "Usage:  " << prog_name << " [--geometry width height] [-p] [-c N] [-f N] [-z N] [-i -c file_name]" << endl
      << "--geometry width height" << endl
      << "  specifies that the window should be width x height" << endl
      << "  The default is " << DEFAULT_WIDTH << 'x' << DEFAULT_HEIGHT << endl
      << "-p" << endl
      << "  Toggles procrastination of drawing pixels until just before updating is done." << endl
      << "  Note that setting this to true has the same effect on the completion time of a script file as real life procrastination has on the completion time of a project." << endl
      << "  Default is false." << endl
//       << "-m" << endl
//       << "  Toggles visibility of the marks.  It starts at " << visible_marks << endl
//       << "  If it ends up false, the marks are invisible, but show up in the operations list file when it is saved." << endl
//       << "  The markings are simply a nop to be used to mark places in the operations file where it might be useful to split it manually." << endl
      << "-r" << endl
      << "  Swaps right and left.  Useful if the templates in the template file are in reverse order." << endl
      << "  (Yeah, I messed up when making one of the template sets)" << endl
      << "-c N" << endl
      << "  Set the critical mass for update area.  Once the area to be updated" << endl
      << "    surpasses this mass, the list of rects to be updated is ignored and" << endl
      << "    the whole screen is updated." << endl
      << "  The default is " << critical_update_area_mass << endl
      << "-f N" << endl
      << "  Sets the frequency at which changes are updated and shown on the screen during file loading." << endl
      << "  N > 2 means after every operation, even if said operation is being multiplied." << endl
      << "  N > 1 means after every operation, with multiplied operations only being counted once." << endl
      << "  N > 0 means after every file." << endl
      << "  N > -1 means after loading all files.  Only makes a difference when loading multiple files." << endl
      << "  N <= -1 means do not update after loading files, making the operations performed by them effectively invisible until something else causes an update." << endl
      << "  Default value for N is" << update_frequency << endl
      << "-z N" << endl
      << "  Sets the starting zoom value.  The last time it is given is the one that it stays at." << endl
      << "-t [file_name]" << endl
      << "  Specifies the line template file to load." << endl
      << "-i -c file_name" << endl
      << "  Specifies to read commands from a file." << endl
      << "  -c specifies that the commands are command lines, not ints." << endl
//       << "  The optional -u option indicates that the input should be read byte by byte, with no formatting." << endl
//       << "    This is useful if you decide to use /dev/random as the input file." << endl
//       << "    If you use the -u option, you must specify the number of bytes to read" << endl
//       << "    from the file, as files like /dev/random do not have an end." << endl
//       << "  If the --lines option is given, it specifies that the file is a list of lines." << endl
//       << "    Otherwise, the file is assumed to be a list of operations." << endl
      << "Some side notes:" << endl
      << "  The --geometry and -m options may be given multiple times." << endl
      << "    For the --geometry tag, the last width and height specified are used." << endl
      << "  The -i option can be specified multiple times.  Each time it is used," << endl
      << "    another entry is added to the list of files to read input from." << endl
      << "    Files are read in the order they are given on the command line." << endl;
  }

/* Bad place for a struct, but it's special purpose and isn't used anywhere else. */
struct file_list_entry_t
{
  bool command_lines;
  string file_name;
};

int main(int argc, char **argv)
  {
    vector<file_list_entry_t> files_to_load_from;
    list<string> template_files;
    int arg_parse_errors= 0;
    template_files.push_back("default_line_templates");
    if(argc > 1)
    {
      for(int n= 1; n < argc; ++n)
      {
        if(!strcmp("--geometry", argv[n]))
        {
          int new_width= 1600, new_height= 1200;
          istringstream foo(argv[++n]);
          if(!(foo >> new_width))
          {
            cerr << "Invalid width:  " << argv[n] << endl;
            --n;
            ++arg_parse_errors;
            continue;
          }
          istringstream bar(argv[++n]);
          if(!(bar >> new_height))
          {
            cerr << "Invalid height:  " << argv[n] << endl;
            --n;
            ++arg_parse_errors;
            continue;
          }
          set_screen_geometry(new_width, new_height);
        }
        else if(!strcmp("-bg", argv[n]))
        {
          set_background_color(hex_str_to_uint32(argv[++n]));
        }
        else if(!strcmp("-p", argv[n]))
        {
          procrastinate_drawing= !procrastinate_drawing;
        }
        else if(!strcmp("-r", argv[n]))
        {
          current_env.swap_r_l= !current_env.swap_r_l;
          cerr << "swap_r_l is " << current_env.swap_r_l << endl;
        }
        else if(!strcmp("-c", argv[n]))
        {
          int new_mass;
          istringstream foo(argv[++n]);
          if(!(foo >> new_mass))
          {
            cerr << "Invalid critical mass:  " << argv[n] << endl;
            --n;
            ++arg_parse_errors;
          }
          else
          {
            critical_update_area_mass= new_mass;
          }
        }
        else if(!strcmp("-f", argv[n]))
        {
          int new_freq;
          istringstream foo(argv[++n]);
          if(!(foo >> new_freq))
          {
            cerr << "Refusing to set invalid frequency:  " << argv[n] << endl;
            --n;
            ++arg_parse_errors;
          }
          else
          {
            update_frequency= new_freq;
          }
        }
        else if(!strcmp("-z", argv[n]))
        {
          unsigned int new_zoom;
          istringstream foo(argv[++n]);
          if(!(foo >> new_zoom))
          {
            cerr << "Refusing to set invalid zoom:  " << argv[n] << endl;
            --n;
            ++arg_parse_errors;
          }
          else
          {
            set_zoom(new_zoom);
          }
        }
        else if(!strcmp("-t", argv[n]))
        {
          template_files.push_back(argv[++n]);
        }
        else if(!strcmp("-i", argv[n]))
        {
//           cerr << "Setting up file list entry." << endl;
          file_list_entry_t tmp_file;
          if(!strcmp("-c", argv[++n]))
          {
            tmp_file.command_lines= true;
          }
          else
          {
            tmp_file.command_lines= false;
          }
          tmp_file.file_name= argv[n];
          if(tmp_file.file_name.find("templates") != string::npos)
          { /* It took me a while, but I finally figured out why it wasn't loading
               the templates sets I specified on the command line.... */
            cerr << tmp_file.file_name << " might be a template file, but was given with the -i option.  Did you mean to use -t?" << endl;
          }
          files_to_load_from.push_back(tmp_file);
//           cerr << "Entry successfully set up and added to list." << endl;
        }
        else if(!strcmp("--help", argv[n]))
        {
          show_usage(argv[0]);
        }
        else
        {
          cerr << "Unrecognized argument:  " << argv[n] << endl;
          ++arg_parse_errors;
          show_usage(argv[0]);
        }
      }
    }
    if(arg_parse_errors > 3)
    {
      cerr << "Too many arg errors, refusing to start." << endl;
      return arg_parse_errors;
    }
    if(init_everything(template_files) != 0)
    {
      /* Hm, I wonder what happens if SDL_Quit() is called after SDL_Init() has
         failed.  The documentation doesn't say.... */
      SDL_Quit();
      return -1;
    }
    for(vector<file_list_entry_t>::iterator n= files_to_load_from.begin(); n != files_to_load_from.end(); ++n)
    { /* No need to check for entries, since for an empty vector begin() == end() */
      cerr << "Loading file " << n->file_name << " as specified in list." << endl;
      if(!n->command_lines)
      {
        current_env.process_int_script(n->file_name);
        /* Done with this entry, so continue to the next. */
        frequency_update_drawn_rects(0);
        cerr << "File loaded successfully." << endl;
      }
      else
      {
        current_env.process_command_script(n->file_name);
        frequency_update_drawn_rects(0);
        cerr << "File loaded successfully." << endl;
      }
    }
    current_env.happy_line_env.draw_current_line();
    frequency_update_drawn_rects(-1);
//     SDL_Delay(2000);
//     SDL_Flip(screen);
//     SDL_Delay(2000);
//     quit= true;
    while(!current_env.quit)
    {
      SDL_Event event;
      while(SDL_PollEvent(&event))
      {
        switch(event.type)
        {
          case SDL_KEYDOWN:
            current_env.process_sdlk(event.key.keysym.sym);
            break;
          case SDL_MOUSEBUTTONDOWN:
            {
              point p(event.button.x, event.button.y);
              string e= int_to_hex_str(color_at_point(p));
              display_error(e);
            }
            break;
          case SDL_ACTIVEEVENT:
            while(!event.active.gain)
            {
              SDL_Delay(500);
              SDL_PollEvent(&event);
            }
            break;
          case SDL_QUIT:
            quit= true;
            break;
          default:
            break;
        }
      }
    }
    SDL_Quit();
    return 0;
  }
